﻿using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  /// Gets the first item from the type stack
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorFirstExpression<T> : StatelessExpression {
    protected VectorFirstExpression() { }
    [StorableConstructor]
    protected VectorFirstExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<IReadOnlyList<T>> vectorStack) {
      return vectorStack.IsEmpty || vectorStack.Top.Count == 0;
    }

    protected void Eval(IPushStack<IReadOnlyList<T>> vectorStack, IPushStack<T> literalStack) {
      var vector = vectorStack.Pop();
      literalStack.Push(vector[0]);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].FIRST",
    "Pushes the first INTEGER of the top INTEGER[].",
    StackTypes.Integer)]
  public class IntegerVectorFirstExpression : VectorFirstExpression<long> {
    public IntegerVectorFirstExpression() { }
    [StorableConstructor]
    protected IntegerVectorFirstExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].FIRST",
    "Pushes the first FLOAT of the top FLOAT[].",
    StackTypes.Float)]
  public class FloatVectorFirstExpression : VectorFirstExpression<double> {
    public FloatVectorFirstExpression() { }
    [StorableConstructor]
    protected FloatVectorFirstExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].FIRST",
    "Pushes the first BOOLEAN of the top BOOLEAN[].",
    StackTypes.Boolean)]
  public class BooleanVectorFirstExpression : VectorFirstExpression<bool> {
    public BooleanVectorFirstExpression() { }
    [StorableConstructor]
    protected BooleanVectorFirstExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].FIRST",
    "Pushes the first BOOLEAN of the top BOOLEAN[].",
    StackTypes.String)]
  public class StringVectorFirstExpression : VectorFirstExpression<string> {
    public StringVectorFirstExpression() { }
    [StorableConstructor]
    protected StringVectorFirstExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}

